/* 
 * Boot Test
 *  version 1.01
 *  dated: 17 Nov 2014
 * 
 *  Purpose:
 *  - The purpose of this app is to create a 10meg disk image, placing an empty
 *    MBR in the first sector, with room for four partition entries.  Then take
 *    a list of four files, which are partition images, and in random order, place
 *    them at the end of the MBR to produce a disk image with four partitions,
 *    updating the MBR partition entries to match.  It will also randomly choose
 *    one of these partitions to set as the active partition.
 * 
 *  - This will be used by the people in the group alt.os.development as a test
 *    app to set their bootsector code to see if it will boot all four partitions
 *    at random tests.
 * 
 *  - Their bootsector code is to be as small as possible, yet still be able to
 *    successfully boot each partition at random.
 * 
 * Use:
 *  - This app needs a target filename, a file that will be created, that is 10 megs
 *    in size, rounded up to the next cylinder boundary, when sectors per track is 63,
 *    and heads per cylinder is 16.
 *  - This file will be generated as a single 512 byte sector size, holding four empty
 *    partition entries.
 *  - It will then require four more filenames, in the form of "part.img", given on
 *    the command line as one filename.  This app will take that command line name and
 *    append a '0', '1', '2', '3' after the name and before the '.img' and then expect
 *    those four files to exist on the host machine:
 *        part0.img, part1.img, part2.img, and part3.img.
 *    This is so there can be numerous testing partitions.
 *  - This app may place a random amount of sectors (0 to 63) between each partition.
 *  - This app will then randomly append each of these partition files to the first
 *    file created, the MBR file.
 *  - This app will then update the partition entries in the MBR sector and randomly
 *    choose one of the four to mark as 'active'
 *  - It is assumed the four partition files are valid partitions, each with a valid
 *    FAT12/16 file system, starting at the first 'sector' of the file.
 *
 * Command Line Usage:
 *   boottest target.img part.img
 *
 *  The 'target.img' is the name of the image file to create.  It will be a minimum
 *   of 10 * 1024 * 1024 bytes in size.
 *  The 'part.img' is the 'short' name of the four partition image files, part0.img,
 *   part1.img, part2.img, and part3.img.
 *
 *  **** For the purpose of this compo, the PARTx.IMG files will be unmodified...
 * 
 */

#ifdef _MSC_VER
  #define _CRT_SECURE_NO_WARNINGS // don't warn about old strcpy(), etc.
#endif

#include <sys/stat.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "boottest.h"

int main(int argc, char *argv[]) {
  struct S_MBR mbr;
  char targfile[FILENAME_MAX];
  char srcfiles[4][FILENAME_MAX];
  FILE *targ, *src;
  char temp[512];
   int i, j, r, sz, len, ch;
   int sequ[4]; // will hold 4 numbers 0 -> 3 randomly sequenced.
  char *p;
  int sys_id[4] = { 1, 1, 4, 4 };
  
  // check and get the command line args.
  if (argc != 3) {
    printf("\nUsage:"
           "\n  boottest target.img part.img");
    return -1;
  }
  
  // save the target filename
  strcpy(targfile, argv[1]);
  
  // create and save the four partition filenames
  strcpy(temp, argv[2]);
  p = strchr(temp, '.');
  if (p == NULL)
    strcat(temp, ".img");
  p = strchr(temp, '.');
  *p = '\0';
  for (i=0; i<4; i++)
    sprintf(srcfiles[i], "%s%i.%s", temp, i, p + 1);
  
  // seed the randomizer
  srand((int) time(NULL));
  
  // create/open the targ file
  if (does_file_exist(targfile)) {
    printf("\nTarget file name '%s' already exists. Overwrite? ", targfile);
    ch = _getche();
    if ((ch == 'Y') || (ch == 'y')) {
      if ((targ = fopen(targfile, "wb")) == NULL) {
        printf("\n Error creating/opening target file: '%s'", targfile);
        return -2;
      }
    } else {
      printf("\n  ...Aborting");
      return -3;
    }
  } else {
    if ((targ = fopen(targfile, "w+b")) == NULL) {
      printf("\n Error creating target file: '%s'", targfile);
      return -2;
    }
  }
  
  // write an empty MBR to the file, starting at offset 0.
  memset(&mbr, 0, sizeof(struct S_MBR));
  fwrite(&mbr, 1, sizeof(struct S_MBR), targ);
  
  // now generate four random numbers between 0 and 3
  generate_sequence(sequ);
  
  for (i=0; i<4; i++) {
    memset(&mbr.part_entry[i], 0, 16);
    if ((src = fopen(srcfiles[sequ[i]], "rb")) == NULL) {
      fclose(targ);
      printf("\n Error opening source file: '%s'", srcfiles[sequ[i]]);
      return -4;
    }
    // randomly put 0 -> 63 sectors before each partition
    // place a random value in this sector so it can not be relied upon.
    r = rand() & 63;
    for (j=0; j<r; j++) {
      memset(temp, rand() & 0xFF, 512);
      fwrite(temp, 1, 512, targ);
    }
    
    mbr.part_entry[i].start_lba = (ftell(targ) / 512);
    mbr.part_entry[i].sys_id = sys_id[sequ[i]];
    calc_chs_from_lba(&mbr.part_entry[i].start, mbr.part_entry[i].start_lba);
    printf("\n Partition %i: '%s': placed at LBA: %i", i, srcfiles[sequ[i]], mbr.part_entry[i].start_lba);
    len = 0;
    do {
      sz = fread(temp, 1, 512, src);
      if (sz > 0) {
        fwrite(temp, 1, 512, targ);  // 512 to remain sector boundry
        len++;
      }
    } while (sz == 512);
    mbr.part_entry[i].size = len;
    calc_chs_from_lba(&mbr.part_entry[i].end, mbr.part_entry[i].start_lba + len - 1);
    fclose(src);
  }
  
  // now pad the image file to the next cylinder boundary
  j = (ftell(targ) / 512);
  if (j % (16 * 63)) {
    r = ((16 * 63) - (j % (16 * 63)));
    printf("\n Padding image with %i sectors.", r);
    while (r--) {
      memset(temp, rand() & 0xFF, 512);
      fwrite(temp, 1, 512, targ);
    }
  }
  
  // now generate four more random numbers between 0 and 3
  generate_sequence(sequ);
  mbr.part_entry[sequ[0]].boot_id = 0x80;
  mbr.sig = 0xAA55;
  printf("\n Marking Partition %i: as active.", sequ[0]);
  
  // copy the MBR code to the first (and at most) 446 bytes of the MBR
  if ((src = fopen("entry.bin", "rb")) == NULL) {
    printf("\n Error opening 'ENTRY.BIN' file.");
    return -5;
  }
  sz = fread(temp, 1, 512, src);
  if (sz > 446) {
    printf("\n Entry.bin file greater than 446 bytes...");
    sz = 446;
  }
  memset(mbr.boot, rand() & 0xFF, 446);  // first write random bytes
  memcpy(mbr.boot, temp, sz);
  fclose(src);
  
  // back up and write the mbr again.
  fseek(targ, 0, SEEK_SET);
  fwrite(&mbr, 1, sizeof(struct S_MBR), targ);
  
  // close the target file handle
  fclose(targ);
  
  return 0;
}

bool does_file_exist(const char *filename) {
  struct stat st;
  int result = stat(filename, &st);
  return (result == 0);
}

// generate four random numbers, 0 -> 3, placing them in sequ[] so that the
//  sequence of sequ[] is not [0][1][2][3], but maybe [3][0][2][1].
void generate_sequence(int *sequ) {
  int i, j;
  
  for (i=0; i<4; i++) {
    sequ[i] = rand() & 3;
    for (j=0; j<i; j++) {
      if (sequ[i] == sequ[j]) {
        i--;
        break;
      }
    }
  }
}

// converts lba to chs and stores in *part
// We assume 63 SPT and 16 HEADS
// Sector   = (LBA mod SPT) + 1
// Head     = (LBA  /  SPT) mod Heads
// Cylinder = (LBA  /  SPT)  /  Heads
void calc_chs_from_lba(struct S_CHS_FIELDS *chs, const bit32u lba) {
  chs->sector   = (bit8u) ((lba % 63) + 1);
  chs->head     = (bit8u) ((lba / 63) % 16);
  chs->cylinder = (bit8u) ((lba / 63) / 16);
}

/*

The following is some information on the way target.img is formatted.
These partx.img files may not in this particular order, they are just
 listed here from part0 to part3.
The information here does not change.

       Creating file:   [part0.img]
           Cylinders:    6
               Sides:    16
       Sectors/Track:    63
       Total Sectors:    6048
                Size:    3.02 (megs)
     Sectors/Cluster:    2
               FAT's:    2
         Sectors/FAT:    9
            FAT size:    12

       Creating file:   [part1.img]
           Cylinders:    4
               Sides:    16
       Sectors/Track:    63
       Total Sectors:    4032
                Size:    2.02 (megs)
     Sectors/Cluster:    1
               FAT's:    1
         Sectors/FAT:    12
            FAT size:    12

       Creating file:   [part2.img]
           Cylinders:    10
               Sides:    16
       Sectors/Track:    63
       Total Sectors:    10080
                Size:    5.04 (megs)
     Sectors/Cluster:    4
               FAT's:    2
         Sectors/FAT:    10
            FAT size:    16

       Creating file:   [part3.img]
           Cylinders:    10
               Sides:    16
       Sectors/Track:    63
       Total Sectors:    10080
                Size:    5.04 (megs)
     Sectors/Cluster:    8
               FAT's:    1
         Sectors/FAT:    5
            FAT size:    16

This is the last time I created target.img.
It will change each time you run the app.
 
 Partition 0: 'part3.img': placed at LBA: 13
 Partition 1: 'part0.img': placed at LBA: 10134
 Partition 2: 'part1.img': placed at LBA: 16194
 Partition 3: 'part2.img': placed at LBA: 20228
 Padding image with 940 sectors.
 Marking Partition 3: as active.

*/

